/*------------------------------------------------------------------------------------------------------------

 Realisiert mit Tiny25
 
 Interner Oszillator mit 8MHz
 
 Layout:
   RESET (ADC0)              1    8    +5V
   Motor-Spannung Neg (PB3)  2     7    PB2 (ADC1): Poti-Spannung
   Motor-Spannung Pos (PB4)  3     6    PB1: Modus (0=Simpel 1=Geregelt)
   GND                       4     5    PB0: FET-Ausgang

 Input:
  PB4  ADC3 - Spannung an Pos. Motorseite (Versorgungsspannung)
  PB3  ADC2 - Spannung an neg. Motorseite (ber'm FET)
  PB2: ADC1 - Poti Spannung
  PB1: Modus (offen= PWM-Regelung ; auf GND= simple PWM-Steuerung)
Output:
  PB0  Ausgang zum FET
*/

#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>

//------------------------------------------------------------------------------------------------------------
int messePotispannung(void){
  ADMUX=1; //ADC1/PB2 Poti-Spannung
  DIDR0 = (1<<ADC1D)|(1<<ADC2D); //Digitale Eingangsstufe abschalten
  ADCSRA = (1<<ADEN) |  //Enable ADC
           (1<<ADSC) |  //Start conversion
           (1<<ADIF) |   //Clear(!) interrupt flag
           (1<<ADPS2)|
           //(1<<ADPS1)|
           (1<<ADPS0);            //ADC Prescaler=32 - gibt 250KHz
           
  //Warte, bis Messung fertig
  while( !( ADCSRA & (1<<ADIF) ) )
  ;  
  ADMUX=6; //ADC2: PB4-PB3   (Motor-Spannung)
  return ADC;
}



//------------------------------------------------------------------------------------------------------------
int messeDiffMotor(void){
  //Motorspannung differentiell messen
  ADMUX=6; // Motor pos an PB4,  neg an PB3,  Gain 1x
  DIDR0 = (1<<ADC1D)|(1<<ADC2D); //Digitale Eingangsstufe abschalten
  ADCSRA = (1<<ADEN) |  //Enable ADC
           (1<<ADSC) |  //Start conversion
           (1<<ADIF) |   //Clear(!) interrupt flag
           (1<<ADPS2)|
           //(1<<ADPS1)|
           (1<<ADPS0);            //ADC Prescaler=32 - gibt 250KHz
           
  ADCSRB = 0;
  //Warte, bis Messung fertig
  while( !( ADCSRA & (1<<ADIF) ) )
  ;  
  ADMUX=1; //ADC1/PB2 Poti-Spannung
  return ADC;
}


//------------------------------------------------------------------------------------------------------------
void delay_25us(void){
static unsigned char i;
  i=10;
  do{
     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");

     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");
     asm volatile("NOP");
  }while (--i);  
}

//------------------------------------------------------------------------------------------------------------
int median(int value){
//ACHTUNG: Darf nur von 1 Stelle im Programm aufgerufen werden!
static int a,b,c, f1,f2,f3, x;
  f1=f2;
  f2=f3;
  f3=value;
  
  a=f1;
  b=f2;
  c=f3;

  if (a>b){
    x=a;
    a=b;
    b=x; 
  } //Jetzt ist a<=b

  if (b>c){
    x=c;
    c=b;
    b=x;
  }// b<=c
  if (a>b){
    x=a;
    a=b;
    b=x; 
  } //Jetzt ist a<=b

  return b;
}

//------------------------------------------------------------------------------------------------------------
int main(void){

  static int ist,soll,i, messungenU, messungenPoti;
  static int tastverhaeltnis;
  static int summeIst;
  static int summeSoll;
  static int delta;
  static int rp,ri;

 
  cli();
  
  DDRB = 1; //PB0 ist Ausgang
  PORTB = 2; //Motor aus; Pull-Up auf Port  PB1
 
  tastverhaeltnis=0;
 
  rp=0;
  ri=0;
  messungenPoti=0;
  messungenU=0;
  summeIst=0;
  summeSoll=0;

  while(1){
  
    summeSoll += messePotispannung();
    messungenPoti++;
    if (32==messungenPoti){
      soll = summeSoll>>5;
      summeSoll=0;
      messungenPoti=0;
    }
    
    messungenU++;
    if (messungenU > 5){
        ist = summeIst / messungenU;
        messungenU=0;
        delta = soll -32 - ist;

        summeIst=0;

        //P-Regler
        if (delta >  512){
          rp=100;
        } else
        if (delta < -512){
         rp=0;
        }else
        rp=50+delta/5;
    
        //I-Regler
        if ((delta >0) && (ri <  1000)) ri += 1 + (delta>>5);
        if ((delta <0) && (ri > -1000)) ri -= 1 - (delta>>5);
    }
    tastverhaeltnis = rp + ri/5;
    //Begrenzung
    if (tastverhaeltnis > 92) tastverhaeltnis=92;;
    if (tastverhaeltnis < 0) tastverhaeltnis =0;

    if (PINB & 2){
      //Geregelter Modus
      summeIst += median(messeDiffMotor()); //Direkt vor dem Einschalten des Motors dessen Generatorspannung messen
      if (soll>31){
        PORTB |= 1; //Motor ein
        for (i=0; i< tastverhaeltnis; i++){
          delay_25us();  
        }  
      }  
      
      PORTB &= ~1; //Motor aus
      for (i=0; i< (100-tastverhaeltnis); i++){
        delay_25us();  
      }  

    }else{
      //Gesteuerter Modus (simpler PWM)
      tastverhaeltnis = soll / 10;
      if (tastverhaeltnis >100) tastverhaeltnis=100;
      if (tastverhaeltnis>0){
        PORTB |= 1; //Motor ein
        for (i=0; i< tastverhaeltnis; i++){
          delay_25us();  
        }  
      }  
      
      PORTB &= ~1; //Motor aus
      for (i=0; i< (100-tastverhaeltnis); i++){
        delay_25us();  
      }  
    }//Modus
  }//While(1)
  return 0;
}
// EOF----------------------------------------------------------------------------------------------------------
